import pytest

from tktl.commands.health import GetGrpcHealthCommand, GetRestHealthCommand
from tktl.core.exceptions.exceptions import APIClientException


def test_health(logged_in_context, test_user_deployed_repos):
    for repo, branch, endpoint in test_user_deployed_repos:
        cmd = GetRestHealthCommand(repository=repo, branch_name=branch, local=False)
        result = cmd.execute(endpoint_name=endpoint)
        assert result.status_code == 200

    rest_repo_name, rest_epo_branch, rest_epo_endpoint = test_user_deployed_repos[0]
    with pytest.raises(APIClientException):
        cmd = GetGrpcHealthCommand(
            repository=rest_repo_name, branch_name=rest_epo_branch, local=False
        )
        cmd.execute(endpoint_name=rest_epo_endpoint)

    grpc_repo_name, grpc_repo_branch, grpc_repo_endpoint = test_user_deployed_repos[-1]
    cmd = GetGrpcHealthCommand(
        repository=grpc_repo_name, branch_name=grpc_repo_branch, local=False
    )
    result = cmd.execute(endpoint_name=grpc_repo_endpoint)
    assert result is True



from tktl import Tktl

from .endpoint_defs.structured import tktl as tktl_structured
from .endpoint_defs.unstructured import tktl as tktl_unstructured

tktl = Tktl()
tktl.endpoints = tktl_structured.endpoints + tktl_unstructured.endpoints



from tests.integration.samples import tktl
from tktl.core.t import EndpointKinds


def test_this():
    assert len(tktl.endpoints) == 7


def test_profiling_enabled():
    for e in tktl.endpoints[0:6]:
        if e.KIND.value == EndpointKinds.CUSTOM.value:
            assert e.profiling_supported is False
        else:
            assert e.profiling_supported is True
    assert tktl.endpoints[-1].profiling_supported is False


def test_app_runs():
    ...



from typing import Union

from pyarrow._flight import FlightClient

from tktl.core.clients.http_client import API
from tktl.core.clients.taktile import TaktileClient
from tktl.core.clients.utils import get_deployment_from_endpoint_and_branch
from tktl.core.config import settings
from tktl.core.loggers import LOG
from tktl.core.t import ServiceT
from tktl.login import login


class Client(object):
    TRANSPORT: str = ServiceT

    def __init__(
        self,
        api_key: str,
        repository_name: str,
        branch_name: str = None,
        local: bool = False,
    ):
        login(api_key)
        self.local = local
        self.taktile_client = TaktileClient(api_url=settings.DEPLOYMENT_API_URL)
        self.repository_name = repository_name
        self.branch_name = branch_name
        self.logger = LOG
        self._location = None
        self._endpoint = None
        self._client = None

    def local_endpoint(self):
        raise NotImplementedError

    @staticmethod
    def format_url(url: str) -> str:
        raise NotImplementedError

    @property
    def client(self) -> Union[API, FlightClient]:
        return self._client

    @property
    def location(self):
        return self._location

    @property
    def endpoint(self):
        return self._endpoint

    @client.setter
    def client(self, client: Union[API, FlightClient]):
        self._client = client

    @location.setter
    def location(self, value):
        self._location = value

    @endpoint.setter
    def endpoint(self, value):
        self._endpoint = value

    def predict(self, *args, **kwargs):
        raise NotImplementedError

    def set_client_and_location(self, location: str, client: Union[API, FlightClient]):
        self.location = location
        self.client = client

    def get_endpoint_and_location(self, endpoint_name: str):
        self.endpoint = endpoint_name
        if self.local:
            location = self.local_endpoint
        else:
            location = self._get_endpoint_location()
            if not location:
                return
            location = self.format_url(location.public_docs_url)
        return location

    def _get_endpoint_location(self):
        endpoint_mapping = self.taktile_client.get_endpoints_for_repository(
            repository=self.repository_name
        )
        filtered_by_user = get_deployment_from_endpoint_and_branch(
            endpoint_mapping=endpoint_mapping,
            endpoint_name=self.endpoint,
            branch_name=self.branch_name,
        )
        if not filtered_by_user:
            return
        return filtered_by_user

    def list_deployments(self):
        raise NotImplementedError

    def get_sample_data(self):
        raise NotImplementedError

    def get_schema(self, *args, **kwargs):
        raise NotImplementedError

    def health(self):
        raise NotImplementedError



from collections.abc import Iterable, Sequence
from functools import singledispatch
from typing import List, Optional, Type, Union

import numpy
import pandas
from pandas.core.dtypes.inference import is_sequence
from pydantic import create_model
from pydantic.main import ModelMetaclass

from tktl.core.exceptions import TaktileSdkError
from tktl.core.serializers.base import CustomDeserializingModelT
from tktl.core.serializers.rest.schema import (
    get_array_model,
    get_custom_model,
    get_dataframe_model,
    get_flat_array_model,
    get_mapping_model,
    get_nested_sequence_model,
    get_series_model,
    get_single_value_model,
)
from tktl.core.serializers.rest.utils import pandas_to_python_type
from tktl.core.t import RestSchemaTypes


@singledispatch
def to_pydantic(
    value, name: str = RestSchemaTypes.SINGLE_VALUE.value, unique_id: str = None
) -> Type[CustomDeserializingModelT]:
    """
    Single dispatch function for serializing values of each of the supported kinds. Each
    of the dispatched functions' implementation is tailored for one of Taktile's supported
    input types. All the implementations return a custom Pydantic model that implements
    a `deserialize` method to be used when deserializing incoming data as a REST request

    Parameters
    ----------
    value : Any
        Input value to be serialized
    name : str
        Name of the model
    unique_id : str
        Unique identifier of the model

    Returns
    -------
    single_value_model : CustomDeserializingModelT
        The Pydantic-based model
    """

    single_value_model = get_single_value_model(value)
    single_value_model.__name__ = get_model_unique_name(name=name, unique_id=unique_id)
    return single_value_model


@to_pydantic.register
def _(
    value: ModelMetaclass,
    name: str = RestSchemaTypes.CUSTOM_MODEL.value,
    unique_id: str = None,
):
    model = get_custom_model(value)
    model.__name__ = get_model_unique_name(name=name, unique_id=unique_id)
    return model


@to_pydantic.register
def _(
    value: pandas.DataFrame,
    name: str = RestSchemaTypes.DATAFRAME.value,
    unique_id: str = None,
    nullable=True,
):
    type_map = {}
    example_map = {}

    for col in value.columns:
        var_type = pandas_to_python_type(value[col])
        if var_type is None:
            types = (type(v) for _, v in value[col].iteritems() if not pandas.isna(v))
            try:
                var_type = next(types)  # type of first item that is not None
            except StopIteration:
                raise TaktileSdkError(
                    f"Column {col} contains all null values and unable to infer its type with .dtype"
                )

        if nullable:
            var_type = Optional[var_type]

        type_map[col] = (var_type, None)

        # Make nxt_val (pandas.isna is recursive so need to check for iterable first)
        vals = (
            v
            for _, v in value[col].iteritems()
            if isinstance(v, Iterable) or not pandas.isna(v)
        )

        try:
            nxt_val = next(vals)
        except StopIteration:
            nxt_val = None  # python built-in version of NaN

        example_map[col] = nxt_val

    index_type = type(next((t for t in value.index)))
    type_map["index"] = (index_type, None)

    base_model = create_model(name, **type_map)  # noqa
    inner_model_unique_id = get_model_unique_name(
        name, unique_id=f"{unique_id}__dt_validate"
    )
    dataframe_model = get_dataframe_model(
        base_model=base_model, example=[example_map], unique_id=inner_model_unique_id
    )
    dataframe_model.__name__ = get_model_unique_name(name=name, unique_id=unique_id)
    return dataframe_model


@to_pydantic.register
def _(
    value: pandas.Series,
    name: str = RestSchemaTypes.SERIES.value,
    unique_id: str = None,
) -> Type[CustomDeserializingModelT]:
    nullable = True if value.isna().sum() > 0 else False
    type_name = "Outcome" if not value.name else value.name
    type_map = {}

    try:
        nxt_val = next(
            v
            for v in value.to_dict().values()
            if isinstance(v, Iterable) or not pandas.isna(v)
        )
    except StopIteration:
        nxt_val = None

    example_map = {type_name: [nxt_val]}

    types = (type(v) for k, v in value.to_dict().items() if not pandas.isna(v))
    var_type = next(types)  # type of first item that is not None
    if var_type == bool:
        var_type = Union[var_type, float]
    if nullable:
        var_type = Optional[var_type]
    type_map[type_name] = (List[var_type], None)
    base_model = create_model(name, **type_map)  # noqa
    series_model = get_series_model(
        series=value, base_model=base_model, example=example_map
    )
    series_model.__name__ = get_model_unique_name(name=name, unique_id=unique_id)
    return series_model


@to_pydantic.register
def _(
    value: numpy.ndarray, name: str = RestSchemaTypes.ARRAY.value, unique_id: str = None
) -> Type[CustomDeserializingModelT]:
    as_list = value.tolist()
    if len(value.shape) == 1:
        first_non_empty = next(v for v in as_list if not pandas.isna(v))
        array_model = get_flat_array_model(
            type(first_non_empty), example=[first_non_empty]
        )
        name = RestSchemaTypes.FLAT_ARRAY.value
        array_model.__name__ = name if not unique_id else f"{name}__{unique_id}"
    else:
        example = replace_nans_and_infs(value[0]).tolist()
        list_model = get_nested_sequence_model(as_list)
        array_model = get_array_model(list_model, example=[example])
        array_model.__name__ = get_model_unique_name(name=name, unique_id=unique_id)
    return array_model


@to_pydantic.register
def _(
    value: Sequence, name: str = RestSchemaTypes.SEQUENCE.value, unique_id: str = None
) -> Type[CustomDeserializingModelT]:
    if isinstance(value, list) or isinstance(value, tuple):
        if len(value) > 0:
            first = value[0]
            if isinstance(first, list):
                sequence_model = get_nested_sequence_model(first)
                sequence_model.__name__ = get_model_unique_name(
                    name=name, unique_id=unique_id
                )

            elif isinstance(first, dict):
                mapping_model = _dict_model(
                    value=first, name=RestSchemaTypes.DICT.value
                )
                sequence_model = get_mapping_model(base_model=mapping_model)
                sequence_model.__name__ = get_model_unique_name(
                    name=name, unique_id=unique_id
                )
            # more specific than numpy's is_sequence implementation
            elif not is_sequence(first):
                sequence_model = get_single_value_model(first)
                sequence_model.__name__ = get_model_unique_name(
                    name=name, unique_id=unique_id
                )
            else:
                raise
        else:
            raise ValueError("Input must be non-empty")
    else:
        raise
    return sequence_model


@to_pydantic.register
def _(
    value: dict, name: str = RestSchemaTypes.DICT.value, unique_id: str = None
) -> Type[CustomDeserializingModelT]:
    mapping_model = _dict_model(value=value, name=name)
    dict_model = get_mapping_model(mapping_model)
    dict_model.__name__ = get_model_unique_name(name=name, unique_id=unique_id)
    return dict_model


def _dict_model(value, name):
    types = {k: (type(v), v) for k, v in value.items()}
    return create_model(name, **types)


def _get_df_example(example_map):
    class Config:
        schema_extra = {"example": example_map}

    return Config


def get_model_unique_name(name: str, unique_id: str) -> str:
    return name if not unique_id else f"{name}__{unique_id}"


def replace_nans_and_infs(array) -> numpy.array:
    as_float = numpy.isnan(array.astype(numpy.float))
    as_inf = numpy.isinf(array.astype(numpy.float))
    replaced_nans = numpy.where(as_float, None, array)
    return numpy.where(as_inf, 0.0, replaced_nans)



import random
from random import randint, sample
from typing import Dict, List, Tuple, Union

import numpy

random.seed(10)
numpy.random.seed(10)


import pandas
from pandas._testing import (
    makeCustomDataframe,
    makeDataFrame,
    makeFloatSeries,
    makeMissingDataframe,
    makeMixedDataFrame,
    makeObjectSeries,
    makeTimeSeries,
)


def make_base_frame(r, c) -> pandas.DataFrame:
    return makeCustomDataframe(r, c, r_idx_type="i")


FLOAT_NANS_FRAME: pandas.DataFrame = makeMissingDataframe()
MIXED_FRAME: pandas.DataFrame = makeMixedDataFrame()


class TestingFrames:
    BASE_FRAME = makeCustomDataframe(
        5, 10, data_gen_f=lambda r, c: randint(1, 100), r_idx_type="i"
    )
    MISSING_FRAME = makeMissingDataframe()

    @staticmethod
    def complete_frame(
        base_frame: pandas.DataFrame, frames: List[pandas.DataFrame]
    ) -> pandas.DataFrame:
        rows = len(base_frame)
        print(base_frame.shape)
        for frame in frames:
            print(frame.shape)
            if len(frame) > rows:
                frame = frame.head(rows)
            else:
                frac = rows // len(frame)
                frame = pandas.concat([frame] * frac).reset_index()
                diff = rows - len(frame)
                frame = pandas.concat([frame, frame.head(diff)]).reset_index()
            base_frame = pandas.concat([base_frame, frame], axis=1, ignore_index=True)
        base_frame.columns = [str(c) for c in base_frame.columns]
        return base_frame


def make_base_array(dims: Tuple[int]) -> numpy.array:
    return numpy.random.randn(*dims)


def mask_array(x, fill_value, pct) -> numpy.array:
    num_values = int(len(x) * pct)
    idx = numpy.arange(len(x))
    numpy.random.shuffle(idx)
    x[idx[:num_values]] = fill_value
    return x


class TestingArrays:
    @classmethod
    def from_frame(cls, df: pandas.DataFrame) -> numpy.array:
        return df.to_numpy()

    @classmethod
    def labels(
        cls,
        length: int,
        as_series: bool = False,
        null_pct: float = 0,
        kind: str = "regression",
        params: Tuple = (0, 1),
        bool_labels: bool = False,
    ) -> Union[numpy.array, pandas.Series]:
        """

        Parameters
        ----------
        bool_labels
        length
        as_series
        null_pct
        kind
        params : Tuple
            u, sigma

        Returns
        -------

        """
        if kind == "regression":
            arr = params[0] + params[1] * numpy.random.random(length)

        else:
            if bool_labels:
                arr = numpy.random.choice(a=[False, True], size=length, p=[0.5, 0.5])
            else:
                arr = numpy.random.uniform(size=length)
        masked = mask_array(arr, None, null_pct)
        return pandas.Series(masked) if as_series else arr


class TestingSequence:
    @classmethod
    def from_frame(cls, df: pandas.DataFrame) -> List[Dict]:
        return df.to_dict(orient="records")

    @classmethod
    def from_array(cls, arr: numpy.array) -> List:
        return arr.tolist()


def random_sample(arr: numpy.array, size: int = 1) -> numpy.array:
    return arr[numpy.random.choice(len(arr), size=size, replace=False)]



import click

from tktl.commands.health import GetGrpcHealthCommand, GetRestHealthCommand
from tktl.core.exceptions.exceptions import APIClientException, CLIError
from tktl.core.managers.auth import AuthConfigManager
from tktl.core.t import ServiceT
from tktl.login import login, logout


@click.command("health", help="Health check your endpoints")
@click.option("-r", "--repo", help="Repository owner & name: owner/repo-name")
@click.option("-e", "--endpoint-name", help="Endpoint name")
@click.option("-b", "--branch-name", help="Branch name", default="master")
@click.option(
    "-s",
    "--service",
    help="Service kind",
    type=click.Choice([ServiceT.REST.value, ServiceT.GRPC.value]),
    default=ServiceT.REST.value,
)
@click.option(
    "-l", "--local", help="Run against local endpoint", is_flag=True, default=False
)
def health(
    repo: str = None,
    endpoint_name: str = None,
    branch_name: str = "master",
    service: str = ServiceT.REST.value,
    local: bool = False,
):
    if (not repo or not endpoint_name) and not local:
        raise CLIError("If not passing repo and endpoint, must pass local = True")

    api_key = AuthConfigManager.get_api_key()
    if local and not api_key:
        login(api_key="NOT SET")
    command = (
        GetRestHealthCommand(repository=repo, branch_name=branch_name, local=local)
        if service == ServiceT.REST.value
        else GetGrpcHealthCommand(repository=repo, branch_name=branch_name, local=local)
    )
    try:
        command.execute(endpoint_name=endpoint_name)
    except (APIClientException, Exception):
        exit(1)
    finally:
        if not api_key:
            logout()



import os

import numpy
import pandas
from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal, assert_series_equal

from tests.integration.samples.endpoints import tktl
from tktl.core.clients.arrow import ArrowFlightClient


def test_fetch_data():
    k = os.environ["TEST_USER_API_KEY"]
    client = ArrowFlightClient(api_key=k, repository_name="test", local=True)
    for endpoint in tktl.endpoints:
        client.authenticate(endpoint_name=f"{endpoint.func.__name__}")
        x, y = client.get_sample_data()
        assert_data_equals(x, endpoint.input_schema.value)
        assert_data_equals(y, endpoint.output_schema.value)


def assert_data_equals(left, right):
    assert type(left) == type(right)
    if isinstance(left, pandas.DataFrame):
        assert_frame_equal(left, right)
    elif isinstance(left, pandas.Series):
        assert_series_equal(left, right, check_names=False)
    elif isinstance(left, numpy.ndarray):
        assert_array_equal(left, right)
    else:
        raise



import random

import numpy

from tktl import Tktl

from ..data.sample_data import (
    FLOAT_NANS_FRAME,
    MIXED_FRAME,
    TestingArrays,
    TestingFrames,
    make_base_frame,
)

tktl = Tktl()
# initialize dictionaries
model = {}
X_test = {}
y_test = {}

random.seed(10)
numpy.random.seed(10)

base_frame = make_base_frame(100, 5)
test_df = TestingFrames.complete_frame(base_frame, [MIXED_FRAME, FLOAT_NANS_FRAME])

test_labels = TestingArrays.labels(
    length=len(test_df), kind="regression", null_pct=0.2, as_series=True
)

test_labels_arr = test_labels.values

test_labels_bool_series = TestingArrays.labels(
    length=len(test_df), kind="binary", null_pct=0.2, bool_labels=True, as_series=True
)

test_labels_bool_arr = test_labels_bool_series.values


@tktl.endpoint(kind="custom", X=test_df, y=test_labels_bool_series)
def list_custom_list(df):
    return test_labels.sample(len(df))



from tktl.commands.health import GetGrpcHealthCommand, GetRestHealthCommand
from tktl.core.config import settings


def test_health(capfd):
    rest_cmd = GetRestHealthCommand(repository="test", branch_name="master", local=True)
    grpc_cmd = GetGrpcHealthCommand(repository="test", branch_name="master", local=True)

    rest_cmd.execute("")
    out, err = capfd.readouterr()
    assert out == f"Service at {settings.LOCAL_REST_ENDPOINT} is up and running\n"

    grpc_cmd.execute("")
    out, err = capfd.readouterr()
    assert out == f"Service at {settings.LOCAL_ARROW_ENDPOINT} is up and running\n"



import json
from typing import List

import click
import yaml
from pydantic import UUID4, BaseModel

from tktl.cli.common import LOG, ClickGroup, get_cmd_shared_options, to_uuid
from tktl.commands import deployments as deployments_commands
from tktl.core.config import settings
from tktl.core.schemas.utils import get_table_from_list_of_schemas
from tktl.login import validate_decorator


@click.group("get", help="Get resources", cls=ClickGroup, **settings.HELP_COLORS_DICT)
def get():
    pass


@get.command("deployments", help="Get deployment resources")
@click.argument("deployment_id", required=False, callback=to_uuid)
@click.option("-c", "--commit", help="SHA commit")
@click.option("-b", "--branch", help="Branch Name")
@click.option("-s", "--status", help="Status in which deployment is")
@get_cmd_shared_options
@validate_decorator
def get_deployment_by_id(
    deployment_id: UUID4,
    commit: str,
    branch: str,
    status: str,
    full: bool,
    all_resources: bool,
    output: str,
):
    command = deployments_commands.GetDeployments()
    result = command.execute(
        deployment_id=deployment_id,
        commit_sha=commit,
        branch_name=branch,
        status_name=status,
        return_all=all_resources,
    )
    keys = ["id", "Branch @ Commit", "Status"] if not full else None
    return produce_output(resources=result, keys=keys, maxwidth=350, output_kind=output)


@get.command("repositories", help="Get repository resources")
@click.argument("repository_id", required=False, callback=to_uuid)
@click.option("-r", "--repo-name", help="Repository name")
@click.option("-o", "--owner", help="Owner of the repository")
@get_cmd_shared_options
@validate_decorator
def get_repository_by_id(
    repository_id: UUID4,
    repo_name: str,
    owner: str,
    full: bool,
    all_resources: bool,
    output: str,
):
    command = deployments_commands.GetRepositories()
    result = command.execute(
        repository_id=repository_id,
        repository_name=repo_name,
        repository_owner=owner,
        return_all=all_resources,
    )
    keys = ["Name"] if not full else None
    return produce_output(resources=result, keys=keys, maxwidth=200, output_kind=output)


@get.command("endpoints", help="Get endpoint resources")
@click.argument("endpoint_id", required=False, callback=to_uuid)
@click.option("-e", "--endpoint-name", help="Endpoint name")
@get_cmd_shared_options
@validate_decorator
def get_endpoint_by_id(
    endpoint_id: UUID4,
    endpoint_name: str,
    all_resources: bool,
    output: str,
    full: bool = True,
):
    command = deployments_commands.GetEndpoints()
    result = command.execute(
        endpoint_id=endpoint_id, endpoint_name=endpoint_name, return_all=all_resources
    )
    return produce_output(resources=result, output_kind=output, keys=None, maxwidth=130)


class GenericResourceList(BaseModel):
    __root__: List


def produce_output(
    resources: List[BaseModel],
    output_kind: str,
    keys: List[str] = None,
    maxwidth: int = None,
):
    if not resources:
        LOG.error("No resources found")
        return
    if output_kind == "stdout":
        table = get_table_from_list_of_schemas(resources, maxwidth=maxwidth, keys=keys)
        LOG.log(table)
    elif output_kind == "json":
        LOG.log(GenericResourceList.parse_obj(resources).json())
    elif output_kind == "yaml":
        LOG.log(yaml.dump_all([json.loads(r.json()) for r in resources]))



from concurrent.futures.thread import ThreadPoolExecutor

from pyarrow import Table
from pyarrow.flight import (
    ClientAuthHandler,
    FlightCancelledError,
    FlightClient,
    FlightDescriptor,
    FlightInfo,
    FlightUnauthenticatedError,
    FlightUnavailableError,
    Ticket,
)

from tktl.core.clients import Client
from tktl.core.config import settings
from tktl.core.exceptions import AuthenticationError
from tktl.core.exceptions.exceptions import APIClientException
from tktl.core.schemas.repository import _format_grpc_url, load_certs
from tktl.core.serializers import deserialize_arrow, serialize_arrow
from tktl.core.t import ServiceT


class ApiKeyClientAuthHandler(ClientAuthHandler):
    """An example implementation of authentication via ApiKey."""

    def __init__(self, api_key: str):
        super(ApiKeyClientAuthHandler, self).__init__()
        self.api_key = api_key

    def authenticate(self, outgoing, incoming):
        outgoing.write(self.api_key)
        self.api_key = incoming.read()

    def get_token(self):
        return self.api_key


class ArrowFlightClient(Client):
    TRANSPORT = ServiceT.GRPC

    def __init__(
        self,
        api_key: str,
        repository_name: str,
        branch_name: str = None,
        local: bool = False,
    ):
        super().__init__(
            api_key=api_key,
            repository_name=repository_name,
            branch_name=branch_name,
            local=local,
        )

    @staticmethod
    def format_url(url: str) -> str:
        return _format_grpc_url(url)

    @property
    def local_endpoint(self):
        return settings.LOCAL_ARROW_ENDPOINT

    def list_deployments(self):
        pass

    def list_commands(self):
        return self.client.list_actions()

    def authenticate(self, endpoint_name: str):
        location = self.get_endpoint_and_location(endpoint_name)
        if not location:
            return
        certs = load_certs()
        client = FlightClient(tls_root_certs=certs, location=location)
        self.logger.trace(f"Performing authentication request against {location}")
        client.authenticate(
            ApiKeyClientAuthHandler(api_key=self.taktile_client.api_key)
        )
        self.set_client_and_location(location=location, client=client)

    def predict(self, inputs):
        table = serialize_arrow(inputs)
        descriptor = self.get_flight_info(command_name=str.encode(self.endpoint))
        writer, reader = self.client.do_exchange(descriptor.descriptor)
        self.logger.trace(f"Initiating prediction request")
        with writer:
            writer.begin(table.schema)
            writer.write_table(table)
            writer.done_writing()
            table = reader.read_all()
            return deserialize_arrow(table)

    def get_sample_data(self):
        if not self.endpoint:
            raise AuthenticationError(
                "Please authenticate against a specific endpoint first"
            )
        x_ticket, y_ticket = (
            str.encode(f"{self.endpoint}__X"),
            str.encode(f"{self.endpoint}__y"),
        )
        return self._get_data(ticket=x_ticket), self._get_data(ticket=y_ticket)

    def _get_data(self, ticket):
        if not self.endpoint:
            raise AuthenticationError(
                "Please authenticate against a specific endpoint first"
            )
        self.logger.trace(f"Fetching sample data from server")
        reader = self.client.do_get(Ticket(ticket=ticket))
        with ThreadPoolExecutor() as executor:
            future = executor.submit(blocking_read, reader)
            return_value = future.result()
            return deserialize_arrow(return_value)

    def get_schema(self):
        if not self.endpoint:
            raise AuthenticationError(
                "Please authenticate against a specific endpoint first"
            )
        info = self.get_flight_info(str.encode(self.endpoint))
        return info.schema

    def get_flight_info(self, command_name: bytes) -> FlightInfo:
        descriptor = FlightDescriptor.for_command(command_name)
        return self.client.get_flight_info(descriptor)

    def health(self):
        try:
            self.logger.trace(f"Connecting to server...")
            self.client.wait_for_available(timeout=1)
        except FlightUnauthenticatedError:
            self.logger.trace(f"Connection successful")
            return True
        except (FlightCancelledError, FlightUnavailableError):
            raise APIClientException(
                detail="Arrow flight is unavailable", status_code=502
            )
        return True


def blocking_read(reader):
    batches = []
    while True:
        try:
            batch, metadata = reader.read_chunk()
            batches.append(batch)
        except StopIteration:
            break
    return Table.from_batches(batches)



import abc

from tktl.core.clients import TaktileClient
from tktl.core.clients.arrow import ArrowFlightClient
from tktl.core.clients.rest import RestClient
from tktl.core.config import settings
from tktl.core.managers.auth import AuthConfigManager
from tktl.core.t import ServiceT


class CommandBase(object):
    def __init__(self, api=None):
        self.api = api


class BaseClientCommand:
    def __init__(
        self,
        kind: ServiceT,
        api_key: str,
        repository: str,
        branch_name: str,
        local: bool = False,
    ):
        self.kind = kind
        self.client_params = {
            "api_key": api_key,
            "repository_name": repository,
            "branch_name": branch_name,
            "local": local,
        }
        self.client = self._get_client()

    def execute(self, *args, **kwargs):
        pass

    def _get_client(self):
        return (
            RestClient(**self.client_params)
            if self.kind == ServiceT.REST.value
            else ArrowFlightClient(**self.client_params)
        )


class BaseApiCommand:
    __metaclass__ = abc.ABCMeta

    def __init__(
        self, api_url=settings.DEPLOYMENT_API_URL,
    ):
        self.api_url = api_url
        self.client = self._get_client()

    @abc.abstractmethod
    def execute(self, *args, **kwargs):
        pass

    def _get_client(self):
        return TaktileClient(api_url=self.api_url)


class BaseDeploymentApiCommand(BaseApiCommand):
    def __init__(self):
        super().__init__(api_url=settings.DEPLOYMENT_API_URL)

    def execute(self, *args, **kwargs):
        raise NotImplementedError


class BaseTaktileApiCommand(BaseApiCommand):
    def __init__(self):
        super().__init__(api_url=settings.TAKTILE_API_URL)

    def execute(self, *args, **kwargs):
        raise NotImplementedError


class BaseRestClientCommand(BaseClientCommand):
    def __init__(self, repository: str, branch_name: str, local: bool):
        super().__init__(
            kind=ServiceT.REST.value,
            api_key=AuthConfigManager.get_api_key(),
            repository=repository,
            branch_name=branch_name,
            local=local,
        )


class BaseGrpcClientCommand(BaseClientCommand):
    def __init__(self, repository: str, branch_name: str, local: bool):
        super().__init__(
            kind=ServiceT.GRPC.value,
            api_key=AuthConfigManager.get_api_key(),
            repository=repository,
            branch_name=branch_name,
            local=local,
        )



from click.testing import CliRunner

from tktl import main
from tktl.core.exceptions.exceptions import CLIError


def test_health(logged_in_context, test_user_deployed_repos):
    """Test the CLI."""
    runner = CliRunner()
    grpc_repo_name, grpc_repo_branch, grpc_repo_endpoint = test_user_deployed_repos[-1]
    for repo, branch, endpoint in test_user_deployed_repos:
        result = runner.invoke(
            main.health, ["-r", repo, "-b", branch, "-e", endpoint, "-s", "rest"]
        )
        assert result.exit_code == 0
    result = runner.invoke(
        main.health,
        [
            "-r",
            grpc_repo_name,
            "-b",
            grpc_repo_branch,
            "-e",
            grpc_repo_endpoint,
            "-s",
            "grpc",
        ],
    )
    assert result.exit_code == 0


def test_health_fail(test_user_deployed_repos):
    repo, branch, endpoint = test_user_deployed_repos[0]
    runner = CliRunner()
    result = runner.invoke(
        main.health, ["-r", repo, "-b", branch, "-e", endpoint, "-s", "rest"]
    )
    # User not logged in
    assert result.exit_code == 1


def test_health_local(logged_in_context):
    runner = CliRunner()
    result = runner.invoke(main.health, ["-s", "rest"])
    assert result.exc_info[0] == CLIError

    result = runner.invoke(main.health, ["-s", "rest", "-l"])
    assert result.exc_info[0] == SystemExit

    # No running local service
    assert result.exit_code == 1
